/*
 * PrettyFaces is an OpenSource JSF library to create bookmarkable URLs.
 * Copyright (C) 2009 - Lincoln Baxter, III <lincoln@ocpsoft.com> This program
 * is free software: you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details. You should have received a copy of the GNU Lesser General
 * Public License along with this program. If not, see the file COPYING.LESSER3
 * or visit the GNU website at <http://www.gnu.org/licenses/>.
 */
package com.ocpsoft.pretty.faces.config;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.StringTokenizer;

import javax.servlet.ServletContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.SAXException;

import com.ocpsoft.pretty.PrettyContext;
import com.ocpsoft.pretty.PrettyException;
import com.ocpsoft.pretty.faces.DynaviewEngine;
import com.ocpsoft.pretty.faces.config.mapping.UrlMapping;
import com.ocpsoft.pretty.faces.config.servlet.WebXmlParser;

/**
 * Implements the Pretty Faces configuration procedure.
 * <p>
 * At application startup time, before any requests are processed, Pretty Faces
 * processes zero or more configuration resources, located according to the
 * following algorithm:
 * </p>
 * <ul>
 * <li>Search for classpath resources named
 * <code>META-INF/pretty-config.xml</code> in the ServletContext resource paths
 * for this web application, and load each as a configuration resource.</li>
 * <li>Check for the existence of a context initialization parameter named
 * <code>com.ocpsoft.pretty.CONFIG_FILES</code>. If it exists, treat it as a
 * comma-delimited list of context relative resource paths (starting with a
 * <code>/</code>), and load each of the specfied resources.</li>
 * <li>Check for the existence of a web application configuration resource named
 * <code>/WEB-INF/pretty-config.xml</code>, and load it if the resource exists.</li>
 * </ul>
 * 
 * @author Aleksei Valikov
 * @author <a href="mailto:lincolnbaxter@gmail.com>Lincoln Baxter, III</a>
 */
public class PrettyConfigurator
{

   private static final Log log = LogFactory.getLog(PrettyConfigurator.class);

   public static final String PRETTY_CONFIG_RESOURCE = "META-INF/pretty-config.xml";
   public static final String DEFAULT_PRETTY_FACES_CONFIG = "/WEB-INF/pretty-config.xml";

   private final ServletContext servletContext;

   private final PrettyConfigParser configParser = new DigesterPrettyConfigParser();
   private final WebXmlParser webXmlParser = new WebXmlParser();
   private final DynaviewEngine dynaview = new DynaviewEngine();

   private PrettyConfig config;

   /**
    * Loads the Pretty Faces configuration.
    */
   public PrettyConfigurator(final ServletContext servletContext)
   {
      this.servletContext = servletContext;
   }

   public void configure()
   {
      try
      {
         final PrettyConfigBuilder builder = new PrettyConfigBuilder();

         feedClassLoaderConfigs(builder);
         feedContextSpecifiedConfig(builder);
         feedWebAppConfig(builder);

         config = builder.build();
         
         validateConfig(config);

         log.trace("Setting config into ServletContext");
         servletContext.setAttribute(PrettyContext.CONFIG_FILES_ATTR, config);
      }
      catch (Exception e)
      {
         throw new PrettyException("Failed to load configuration.", e);
      }
   }

   private void validateConfig(PrettyConfig config)
   {
      for (UrlMapping m : config.getMappings())
      {
         m.getPatternParser();
      }
      
   }

   public String getFacesDynaViewId()
   {
      try
      {
         webXmlParser.parse(servletContext);
         return dynaview.buildDynaViewId(webXmlParser.getFacesMapping());
      }
      catch (Exception e)
      {
         throw new PrettyException("Could not retrieve DynaViewId.", e);
      }
   }

   private void feedClassLoaderConfigs(final PrettyConfigBuilder builder) throws IOException, SAXException
   {
      final Enumeration<URL> urls = getResourceLoader().getResources(PRETTY_CONFIG_RESOURCE);
      if (urls != null)
      {
         while (urls.hasMoreElements())
         {
            final URL url = urls.nextElement();
            if (url != null)
            {
               InputStream is = null;
               try
               {
                  is = openStream(url);
                  configParser.parse(builder, is);
               }
               finally
               {
                  if (is != null)
                  {
                     try
                     {
                        is.close();
                     }
                     catch (IOException ignored)
                     {

                     }
                  }

               }
            }
         }
      }
   }

   private void feedContextSpecifiedConfig(final PrettyConfigBuilder builder) throws IOException, SAXException
   {
      final List<String> configFilesList = getConfigFilesList();
      for (final String systemId : configFilesList)
      {
         final InputStream is = servletContext.getResourceAsStream(systemId);
         if (is == null)
         {
            log.error("Pretty Faces config resource [" + systemId + "] not found.");
            continue;
         }

         log.trace("Reading config [" + systemId + "].");

         try
         {
            configParser.parse(builder, is);
         }
         finally
         {
            try
            {
               is.close();
            }
            catch (IOException ignored)
            {

            }
         }
      }
   }

   private List<String> getConfigFilesList()
   {
      final String configFiles = servletContext.getInitParameter(PrettyContext.CONFIG_FILES_ATTR);
      final List<String> configFilesList = new ArrayList<String>();
      if (configFiles != null)
      {
         final StringTokenizer st = new StringTokenizer(configFiles, ",", false);
         while (st.hasMoreTokens())
         {
            final String systemId = st.nextToken().trim();

            if (DEFAULT_PRETTY_FACES_CONFIG.equals(systemId))
            {
               log.warn("The file [" + DEFAULT_PRETTY_FACES_CONFIG + "] has been specified in the [" + PrettyContext.CONFIG_FILES_ATTR + "] context parameter of " + "the deployment descriptor. This will automatically be removed, " + " otherwise, it would be loaded twice.");
            }
            else
            {
               configFilesList.add(systemId);
            }
         }
      }
      return configFilesList;
   }

   private void feedWebAppConfig(final PrettyConfigBuilder builder) throws IOException, SAXException
   {

      final InputStream is = servletContext.getResourceAsStream(DEFAULT_PRETTY_FACES_CONFIG);
      if (is != null)
      {
         log.trace("Reading config [" + DEFAULT_PRETTY_FACES_CONFIG + "].");

         try
         {
            configParser.parse(builder, is);
         }
         finally
         {
            try
            {
               is.close();
            }
            catch (IOException ignored)
            {

            }
         }
      }
   }

   /**
    * Return class loader to be used to resolve resources.
    * 
    * @return Class loader to be used to resolve resources
    */
   protected ClassLoader getResourceLoader()
   {
      final ClassLoader resourceLoader = Thread.currentThread().getContextClassLoader();
      return resourceLoader;
   }

   /**
    * Opens an input stream for the given URL.
    * 
    * @param url target URL.
    * @return Opened input stream.
    * @throws IOException If connection could not be opened.
    */
   protected InputStream openStream(final URL url) throws IOException
   {
      final URLConnection connection = url.openConnection();
      connection.setUseCaches(false);
      return connection.getInputStream();
   }

   public PrettyConfig getConfig()
   {
      return config;
   }

}
